/*

Package fe contains the front-end (directly user facing) code for
autoradio.

The front-end code has the purpose of directing user traffic where we
want it: that is, on a node that is alive and (possibly) not
overloaded. We do this at two different levels, DNS and HTTP, with two
slightly different targets: the former is focused on availability,
while the latter attempts to evenly distribute resource usage.

Request Flow

DNS is used to provide clients with at least one address of an active
server: we can simply return a subset of the active nodes on every
request, maximizing the chances that at least one of them will be
active over a longer period of time.

HTTP requests are istantaneous, and clients usually do not implement a
retry behavior, so we must point the user at a single, active node on
every request. There are two different policies, depending on the type
of the request:

- SOURCE requests must always reach the current master node. Since
redirects tend to confuse streaming sources, which might have very
simple HTTP implementations, we simply proxy the stream to the master
node.

- Listener requests must be routed to an available relay, taking
utilization into account (be it in terms of bandwidth, cpu usage, or
more). The fact that the DNS layer returns multiple addresses provides
already a very rough form of load balancing, but for accurate
bandwidth planning we can't just rely on clients' cooperation. So when
a client requests a stream using its public URL we need to serve a
redirect to the desired node, computed according to the load balancing
policy. This is currently done by serving a M3U file pointing directly
at the target node's icecast daemon (but this may lock clients to that
specific target node on failure... client reconnection policies still
need some investigation).

URLs

Autoradio uses a few different types of URLs when interacting with
clients, with different scopes and purposes. For instance, with a
mount named "/test.ogg", we would have:

The main public URL for the stream:

 http://stream.${DOMAIN}/test.ogg.

It's a permanent URL that will always be valid. Clients will use
this URL to listen to the stream. M3U files must contain this URL.

The response from the previous URL will be a redirect to a specific IP:

 http://${IP}/_stream/test.ogg

This is just a proxy to the Icecast daemon running on that machine.
The chosen IP is either one chosen by the load balancing algorithm, if
the request is a GET from a client, or the current master in case of a
SOURCE request.

Assumptions

There are some fundamental assumptions in the traffic model presented
here, which are not entirely true in the real world:

1) Clients are able to follow HTTP redirects:

Browsers apparently do (though they try really hard to cache the
redirects, which we do not want), but a number of old streaming radio
clients probably don't. Some old streaming audio libraries have really
bare-bones HTTP client implementations. This problem appears to be
even worse for sources.

2) Clients (or users) will retry if the connection is interrupted:

Autoradio servers will usually drop connections on most
cluster-level events, so it is expected that clients will retry and
reconnect. Most clients don't do anything like that unfortunately,
but perhaps users will (if the stream suddenly stops).

*/
package fe
